︠2746fcfb-4eaa-4fa4-8818-2477f4aac4eci︠
%html
<p><em><strong>Fl&aacute;vio Code&ccedil;o Coelho</strong></em> - (<a href="http://pyinsci.blogspot.com" target="_blank">http://pyinsci.blogspot.com</a>)</p>
<p>license: Creative Commons non-commercial share-alike</p>
<hr />
<p>This tutorial is concerned with the exploration of the problem of sampling from high dimensional probability distributions. To run this worksheet you will need to install an additional package into Sage with this command: <strong><span style="font-family: andale mono,times;">./sage -python -m easy_install -U BIP</span></strong></p>
<p>To illustrate the problem lets start with 2 dimensions, and a bivariate normal distribution.</p>
<p>&nbsp;</p>

︡073d9f50-2d7d-48e7-b45c-3f30e8ba7df6︡{"html": "<p><em><strong>Fl&aacute;vio Code&ccedil;o Coelho</strong></em> - (<a href=\"http://pyinsci.blogspot.com\" target=\"_blank\">http://pyinsci.blogspot.com</a>)</p>\n<p>license: Creative Commons non-commercial share-alike</p>\n<hr />\n<p>This tutorial is concerned with the exploration of the problem of sampling from high dimensional probability distributions. To run this worksheet you will need to install an additional package into Sage with this command: <strong><span style=\"font-family: andale mono,times;\">./sage -python -m easy_install -U BIP</span></strong></p>\n<p>To illustrate the problem lets start with 2 dimensions, and a bivariate normal distribution.</p>\n<p>&nbsp;</p>"}︡
︠c5e23742-25e6-4fe9-921b-d3d0d41b99e6a︠
%auto
import numpy as np
from numpy.random import normal, multivariate_normal
import scipy.stats as st
import time
import pylab as P
from matplotlib import mlab
from scipy.special import gamma
from BIP.Bayes import lhs, like
︡e0ee231a-3676-4a6f-bf3f-2d39a9430f08︡︡
︠e4a7ff0e-632d-4314-84d5-01a63425a917i︠
%html
<p>After some basic imports, let's look at the shape of the bivariate normal with means $\mu_1$ and $\mu_2$ equal to zero, standard deviations, $\sigma_1$ and $\sigma_2$, equal to 1, and no covariance.</p>

︡1c0fbe36-4989-4bac-8e9d-c53a3ce6aef5︡{"html": "<p>After some basic imports, let's look at the shape of the bivariate normal with means $\\mu_1$ and $\\mu_2$ equal to zero, standard deviations, $\\sigma_1$ and $\\sigma_2$, equal to 1, and no covariance.</p>"}︡
︠65975488-fd4b-4725-955d-b9ed309e4f58a︠
%auto
x,y = np.meshgrid(np.arange(-2.1,2.1,.1),np.arange(-2.1,2.1,.1))
z = mlab.bivariate_normal(x,y,1,1,0,0,0)
P.clf()
P.contour(x,y,z)
P.imshow(z,extent=[-2.1,2.1,-2.1,2.1])
P.savefig('binormal.png')
︡906eb6e8-82f7-458d-a0f4-0d23f27fdb75︡︡
︠634c92c0-79e6-4620-b8d6-c998fc5055b4i︠
%html
<p>We built this bivariate normal contour plot by uniformly sampling along x, y. If we happened to be interested only in samples within 2 standard deviations from the mean, we can easily conclude, by looking at the figure above, that the all the samples coming from the dark blue regions of the picture, would represent wasted sampling effort. So the proportion of wasted effort here would be aproximately  the difference between the area of the figure (square) and the area of largest circle we see in the figure divided by the area of the square:</p>

︡85e7e94c-794d-474b-913c-716043ce6ec3︡{"html": "<p>We built this bivariate normal contour plot by uniformly sampling along x, y. If we happened to be interested only in samples within 2 standard deviations from the mean, we can easily conclude, by looking at the figure above, that the all the samples coming from the dark blue regions of the picture, would represent wasted sampling effort. So the proportion of wasted effort here would be aproximately  the difference between the area of the figure (square) and the area of largest circle we see in the figure divided by the area of the square:</p>"}︡
︠ce713c87-fd10-4452-a345-84470f05659b︠
(16-pi*2**2)/16
︡622e395f-7225-4e69-89ee-77662d428ce3︡{"html": "<span class=\"math\">\\newcommand{\\Bold}[1]{\\mathbf{#1}}-\\frac{1}{4} \\, \\pi + 1</span>"}︡
︠45c6a794-c229-4cf6-bca3-ba514d360ef0i︠
%html
<p>or, in numerical terms:</p>

︡78ba8c93-c25a-4fd0-b45b-59ac418db3fc︡{"html": "<p>or, in numerical terms:</p>"}︡
︠328d5fe4-29e9-4966-be20-fd6f177c80c1︠
n((16-pi*2**2)/16)
︡9983e113-41f0-4d8d-8f9c-ba589635e8c3︡{"html": "<span class=\"math\">\\newcommand{\\Bold}[1]{\\mathbf{#1}}0.214601836602552</span>"}︡
︠c025ac02-5074-45d0-9b83-cb9380d8aba5i︠
%html
<p>Approximately 21% of wasted sampling effort.</p>
<p>But what happens to that ratio as we increase the dimensionality of the target distribution?</p>
<p>the volume of the unit hypersphere is given by the following equation:</p>
<p>$$V_n=\frac{\pi^{n/2}}{\Gamma(n/2+1)}*R^2$$</p>
<p>where n is its dimensionality, $R$ is its radius and $\Gamma$ is the gamma function.</p>
<p>The volume of the circunscribing hypercube, which has edge of length 2, is given by $2^n$. With these formulas, we can easily calculate the efficiency of sampling from an n-dimensional gaussian.</p>

︡b05d2c7e-5daf-41ad-bd1f-4d71741189ce︡{"html": "<p>Approximately 21% of wasted sampling effort.</p>\n<p>But what happens to that ratio as we increase the dimensionality of the target distribution?</p>\n<p>the volume of the unit hypersphere is given by the following equation:</p>\n<p>$$V_n=\\frac{\\pi^{n/2}}{\\Gamma(n/2+1)}*R^2$$</p>\n<p>where n is its dimensionality, $R$ is its radius and $\\Gamma$ is the gamma function.</p>\n<p>The volume of the circunscribing hypercube, which has edge of length 2, is given by $2^n$. With these formulas, we can easily calculate the efficiency of sampling from an n-dimensional gaussian.</p>"}︡
︠2d240283-133e-457a-96c1-be5579fe8793a︠
%auto
w = [N(pi**(d/2)/gamma(d/2+1)/2**d) for d in range(20)]

@interact
def waste(nd=slider(0,20,1,label="number of dimensions",default=2)):
    """
    Calculates the ratio of the volume of the unit hypersphere to its circumscribed hypercube
    """
    P.clf()
    P.plot(range(20),w);P.xlabel("Dimensionality");P.ylabel("Fraction of samples inside the hypersphere")
    P.scatter(nd,w[nd],color='r');P.grid()
    P.savefig('dim.png')
︡840a8073-c505-4b95-b22f-6cd2be324d37︡{"html": "<!--notruncate--><div padding=6 id=\"div-interact-19\"> <table width=800px height=20px bgcolor=\"#c5c5c5\"\n                 cellpadding=15><tr><td bgcolor=\"#f9f9f9\" valign=top align=left><table><tr><td align=right><font color=\"black\">number of dimensions&nbsp;</font></td><td><table><tr><td>\n        <div id=\"slider-nd-19\" style=\"margin:0px; margin-left: 1.0em; margin-right: 1.0em; width: 15.0em;\"></div>\n        </td><td><font color=\"black\" id=\"slider-nd-19-lbl\"></font></td></tr></table><script>(function(){ var values = [\"0\",\"1\",\"2\",\"3\",\"4\",\"5\",\"6\",\"7\",\"8\",\"9\",\"10\",\"11\",\"12\",\"13\",\"14\",\"15\",\"16\",\"17\",\"18\",\"19\",\"20\"]; setTimeout(function() {\n    $('#slider-nd-19').slider({\n        step: 1, min: 0, max: 20, value: 2,\n        change: function (e,ui) { var position = ui.value; if(values!=null) $('#slider-nd-19-lbl').text(values[position]); interact(19, '_interact_.update(\\'19\\', \\'nd\\', 1, _interact_.standard_b64decode(\\''+encode64(position)+'\\'), globals()); _interact_.recompute(\\'19\\');'); },\n        slide: function(e,ui) { if(values!=null) $('#slider-nd-19-lbl').text(values[ui.value]); }\n    });\n    if(values != null) $('#slider-nd-19-lbl').text(values[$('#slider-nd-19').slider('value')]);\n    }, 1); })();</script></td></tr>\n</table><div id=\"cell-interact-19\"><?__SAGE__START>\n        <table border=0 bgcolor=\"white\" width=100% height=100%>\n        <tr><td bgcolor=\"white\" align=left valign=top><pre><?__SAGE__TEXT></pre></td></tr>\n        <tr><td  align=left valign=top><?__SAGE__HTML></td></tr>\n        </table><?__SAGE__END></div></td>\n                 </tr></table></div>\n                 "}︡
︠7eff33f9-5798-4d81-8554-b40cfd4c50cbi︠
%html
<p>As we can see, sampling in multidimensional spaces quickly becomes prohibitively inneficient.</p>
<p>However these results assume uniform sampling along the axes. What if there was a way to concentrate our sampling effort to the higher probability areas? The good news is that there are ways.</p>
<p>The most basic way to do random sampling from a probability distribution works by some form of rejection sampling. This works basically by sampling uniformly from the support of the distribution (in the case of a normal distribution, from $-\inf$ to $\inf$), and retaining them according to their probability as given by the PDF. Even though the final sample will have less points in the tails, the underlying sampling algorithm still suffers from the curse of dimensionality and is going to be very inneficcient at high dimensionalities.</p>
<p>However more clever implementations of random number generators have succesfully avoided the curse of dimensionality. Let's see how Numpy multivariate normal fares in this category:</p>

︡77678806-b6e9-4c71-a74a-a6795122c280︡{"html": "<p>As we can see, sampling in multidimensional spaces quickly becomes prohibitively inneficient.</p>\n<p>However these results assume uniform sampling along the axes. What if there was a way to concentrate our sampling effort to the higher probability areas? The good news is that there are ways.</p>\n<p>The most basic way to do random sampling from a probability distribution works by some form of rejection sampling. This works basically by sampling uniformly from the support of the distribution (in the case of a normal distribution, from $-\\inf$ to $\\inf$), and retaining them according to their probability as given by the PDF. Even though the final sample will have less points in the tails, the underlying sampling algorithm still suffers from the curse of dimensionality and is going to be very inneficcient at high dimensionalities.</p>\n<p>However more clever implementations of random number generators have succesfully avoided the curse of dimensionality. Let's see how Numpy multivariate normal fares in this category:</p>"}︡
︠142f3966-c0a4-4739-804b-418c3295fb7f︠
%python
def time_sampling(n):
    t0 = time.time()
    if n == 1:
        normal(size=10000)
    else:
        multivariate_normal(np.zeros(n),np.identity(n),size=10000)
    return time.time()-t0
P.clf()  
P.plot(range(1,30),[time_sampling(i) for i in range(1,30)],'o')
P.xlabel('dimensions')
P.ylabel('time to generate 10000 samples(s)')
P.savefig('mvsampleeff.png')
︡0b4e2446-d357-4cb5-9d3c-71b98e351015︡︡
︠e2d2c60b-a9b1-41bf-a3eb-f43d8a4424a4i︠
%html
<p>As we can see from the time cost to generates a 10000 valid samples increases linearly with dimensionality, and not exponentially as the curse of dimensionality might predict.</p>
<p>But loss of efficiency is not the only ill-effects of the curse, Whenever we sample from a PDF, we want our sample histogram to resemble the PDF we're sampling from. This can be a big problem especially for distributions with complex shapes (e.g. multimodal) in high dimensions.</p>
<p>To addres this particular issue we can resort to a clever sampling technique know as Latin Hypercube Sampling (LHS). This technique basically aims to guarantee that our samples are "well" spread out, guaranteeing better adherence of the sample distribution to the original PDF.</p>
<p>The BIP package has an implementation of&nbsp; <a href="http://en.wikipedia.org/wiki/Latin_hypercube_sampling" target="_blank">LHS</a> in Python which integrates well with Numpy's and scipy's RNGs. Like the <strong><span style="font-family: courier new,courier;">multivariate_normal</span></strong>, from Numpy, it can also take into account the correlation structure between the variables of our joint probability distribution, and also allows for each variable have <strong>any</strong> distribution type supported by Numpy. One difference to be aware of, is that it takes the <strong>correlation matrix</strong> as an argument and not the covariance.</p>
<p>lets take a look at it:</p>

︡fd0cd089-8d69-4d94-984d-44530aa686f5︡{"html": "<p>As we can see from the time cost to generates a 10000 valid samples increases linearly with dimensionality, and not exponentially as the curse of dimensionality might predict.</p>\n<p>But loss of efficiency is not the only ill-effects of the curse, Whenever we sample from a PDF, we want our sample histogram to resemble the PDF we're sampling from. This can be a big problem especially for distributions with complex shapes (e.g. multimodal) in high dimensions.</p>\n<p>To addres this particular issue we can resort to a clever sampling technique know as Latin Hypercube Sampling (LHS). This technique basically aims to guarantee that our samples are \"well\" spread out, guaranteeing better adherence of the sample distribution to the original PDF.</p>\n<p>The BIP package has an implementation of&nbsp; <a href=\"http://en.wikipedia.org/wiki/Latin_hypercube_sampling\" target=\"_blank\">LHS</a> in Python which integrates well with Numpy's and scipy's RNGs. Like the <strong><span style=\"font-family: courier new,courier;\">multivariate_normal</span></strong>, from Numpy, it can also take into account the correlation structure between the variables of our joint probability distribution, and also allows for each variable have <strong>any</strong> distribution type supported by Numpy. One difference to be aware of, is that it takes the <strong>correlation matrix</strong> as an argument and not the covariance.</p>\n<p>lets take a look at it:</p>"}︡
︠11830365-d2a2-46a4-b963-c283138fcfd4︠
cm = np.array([[1,.8],[.8,1]])
c=lhs.lhs([st.norm,st.beta], [(50,1),(10,2)],2000,False, cm)
r,p = st.spearmanr(c[0],c[1])
print "Correlation: %1.4f, p-value: %s"%(r,p)
P.clf()
P.scatter(c[0],c[1]);P.xlabel("N(50,1)");P.ylabel("Beta(10,2)")
P.savefig('lhs.png')
︡4f8e1a6b-bc3f-4c30-93e0-86d69b71dd72︡{"stdout": "Correlation: 0.7909, p-value: 0.0"}︡
︠f8d1b2f6-4f1d-492d-9174-aeb788619ba0i︠
%html
<p>The figure above contains a scatter plot of a bivariate Normal-Beta distribution havin a correlation of 0.8 between its variables.</p>
<p>Now that we have seen that it works, Let's switch back to our discussion about the Curse, and how LHS can help us to avoid its ill effects. To check if LHS sampling can&nbsp; improve the quality of our sampling, let's go back to multivariate normals, for simplicity sake.</p>
<p>We need a little more code first:</p>

︡84566d9b-2f7d-4d89-bc89-b346a2ad1e3e︡{"html": "<p>The figure above contains a scatter plot of a bivariate Normal-Beta distribution havin a correlation of 0.8 between its variables.</p>\n<p>Now that we have seen that it works, Let's switch back to our discussion about the Curse, and how LHS can help us to avoid its ill effects. To check if LHS sampling can&nbsp; improve the quality of our sampling, let's go back to multivariate normals, for simplicity sake.</p>\n<p>We need a little more code first:</p>"}︡
︠1cb186fc-9f40-44a7-8197-2f4eef01149f︠
def get_mvn_samples(n,siz=100):
    siz //=n #makes the total number of samples be siz, not siz*n
    ls = np.array(lhs.lhs([st.norm]*n,[(0,1)]*n,siz,False,np.identity(n))).T
    rs = multivariate_normal(np.zeros(n),np.identity(n),size=siz)
    ls.shape = (siz,n)
    return rs,ls

def normality_test(n,siz):
    rs,ls = get_mvn_samples(n,siz)
    rsl = np.sum([like.Normal(rs[:,i],0,1) for i in range(n)])
    lsl = np.sum([like.Normal(ls[:,i],0,1) for i in range(n)])
    return np.exp(rsl-lsl)
︡ed1febf1-a032-43d7-853f-4beb6148ca50︡︡
︠5dab8c39-b9f4-43bb-81a7-ee0df8227c14i︠
%html
<p>The first function above, generates samples from multivariate normal distributions of a given size and dimensianality. It returns two samples: one samples randomly, and the other sampled through LHS.</p>
<p>The second function above calculates the log-likelihood of the samplesunder a normal model. It returns the likelihood ratio ($\frac{\Lambda{random}}{\Lambda_{LHS}}$), which for log-likelihoods, translates into the difference between them.</p>
<p>So let's look at some results:</p>

︡fd1d638a-cb53-4a93-afc3-658c295e8aa6︡{"html": "<p>The first function above, generates samples from multivariate normal distributions of a given size and dimensianality. It returns two samples: one samples randomly, and the other sampled through LHS.</p>\n<p>The second function above calculates the log-likelihood of the samplesunder a normal model. It returns the likelihood ratio ($\\frac{\\Lambda{random}}{\\Lambda_{LHS}}$), which for log-likelihoods, translates into the difference between them.</p>\n<p>So let's look at some results:</p>"}︡
︠58d77fe9-13bf-4b2b-84cb-7cf01c57e4f2︠
%python
P.clf()
P.plot(range(1,60),[normality_test(i,4000) for i in xrange(1,60)],'o')
P.title('Sample size = 4000');P.yscale('log')
P.xlabel('dimensions')
P.ylabel('Likelihood ratio of samples N(0,1)')
P.savefig('lik_dims.png')
︡5c939b19-311f-4ac5-a712-5734e4342782︡︡
︠6ee2ab4f-0efd-4531-abe0-8d227628aa90i︠
%html
<p>The plot above makes it clear that with in higher dimensionality, LHS sampling better resembles a multivariate Normal distribution while at lower dimensions it does not give us much advantage over random sampling even at mildly high dimensionalities. In the plot above the smaller the number the better the LHS compared to random sampling.</p>

︡b472901d-7225-4fbd-8f01-fa95743e1bcb︡{"html": "<p>The plot above makes it clear that with in higher dimensionality, LHS sampling better resembles a multivariate Normal distribution while at lower dimensions it does not give us much advantage over random sampling even at mildly high dimensionalities. In the plot above the smaller the number the better the LHS compared to random sampling.</p>"}︡
︠3a3184c3-ca61-4c1a-8f22-80f26dcd7813︠
%python
P.clf()
P.plot(range(100,500),[normality_test(5,i) for i in xrange(100,500)],'o')
P.plot(range(100,500),ones(400),'r:')#LR=1 line
P.xlabel('sample size - 5D MVN');P.yscale("log")
P.ylabel('Likelihood ratio of samples N(0,1)')
P.savefig('lik_ssize.png')
︡ae152fa2-f6c8-4ddc-83ef-ea8875d7afaf︡︡
︠d6ff83c8-d217-43a2-843b-2af40693e407i︠
%html
<p>The plot above indicates that sample size alone, when you fix dimensionality (5D, in this case), does not seem to affect how much improvement in the normality of the sample provided&nbsp; by the LHS method. LHS is still beter at just about all sample sizes tested here (LR &lt; 1).</p>

︡b33cea6d-d603-4c8d-83da-84484b72baee︡{"html": "<p>The plot above indicates that sample size alone, when you fix dimensionality (5D, in this case), does not seem to affect how much improvement in the normality of the sample provided&nbsp; by the LHS method. LHS is still beter at just about all sample sizes tested here (LR &lt; 1).</p>"}︡
